"""
Principles of implementing scheduler:
- clicking on "Run" for a job should cause no delay

- there should be a sensible and clear priority structure over
  which actions are possible when a job is in various statuses
  (e.g. it should not be possible to run a job when it is in
  the process of being spawned, or is already running)
- the above principle should extend to such rules as: if users
  click "Run" for the same job in quick succession, it should
  only run once

- there should be sensible logic for trying to assign a job
  to a host, and how to retry running the job if no appropriate
  host is available first time (and this should be non-blocking,
  it shouldn't block scheduler's core processing loop, however
  that is implemented

- logic around resetting and rescheduling jobs should be clear
  (e.g. if we want to have a job with multiple schedules)

- ideally all actions should be easily trackable, so for example
  when we click on "Run", we might receive a 200 indicating that
  the job was added to the queue successfully, but it would be
  nice to be able to track its progress through the various stages
  of spawn/run somehow (maybe this isn't necessary since all jobs
  have statuses now, these statuses are clearly visible)
"""

import os
import queue
import subprocess
import threading
import time
from collections import defaultdict
from multiprocessing import current_process
from typing import Dict, List, Union

import schedule
from pebble import concurrent

from jobdec.action import (
    Kill,
    RerunAfterDelay,
    Reset,
    RetryAfterDelay,
    Run,
    RunFromSchedule,
    SetCompletedFailure,
    SetCompletedSuccess,
    Spawn,
    StopProcessingQueueAction,
)
from jobdec.file_utils import full_path
from jobdec.graph import simple_jobs_to_networkx_graph
from jobdec.job_decorator import JobsCollection
from jobdec.scheduled_job import JobStatuses, ScheduledJob
from jobdec.utils import logger


class Scheduler:
    def __init__(
        self,
        name: str,
        jobs: Union[List[ScheduledJob], Dict[str, ScheduledJob]],
        retry_delays: list = None,
        hosts: dict = None,
        run_threads: bool = True,
    ):
        """
        :param name:
          A name for the scheduler object. If you only use one
          Scheduler object for all jobs in your repo, simply use
          the repository name.
          This name is used to differentiate between Scheduler
          objects, in particular it is used as the folder where
          logs for the jobs are stored. This means that when
          multiple Scheduler objects are used on the same machine
          (possibly within different repositories), log files are
          kept separate.

        :param retry_delays:
          A list containing the delays (in seconds) to use when
          retrying spawning a job, due to not finding a suitable
          host. For example, to attempt three retries, after
          10, then 20, then 60 seconds, simply pass in a list
          [10, 20, 60]. To not implement any sort of retry logic,
          pass in an empty list [], or None.

        :param run_threads:
          Should the worker threads which process the queue
          and scheduled jobs be set running?
          In some unit tests it is convenient to not run
          these threads and do things manually, one step
          at a time.
        """
        # We assume the local machine has 10GB
        # of RAM available if not told otherwise.
        self.name = name
        if isinstance(jobs, dict):
            self.job_lookup = dict(jobs)
            self.jobs = list(jobs.values())
        else:
            self.jobs = list(jobs)
            self.job_lookup = {t.name: t for t in jobs}

        bad_names = [
            (k, job.name) for k, job in self.job_lookup.items() if k != job.name
        ]
        assert not bad_names, f"Job names must match dictionary keys: {bad_names}"

        self.hosts = hosts or {"localhost": 10}
        self.retry_delays = retry_delays or []

        self.graph = simple_jobs_to_networkx_graph(self.jobs)

        self.hosts_lock = threading.Lock()

        # Using a queue which is an attribute of the Scheduler
        # instance makes unit testing a lot easier.
        self.queue = queue.Queue()

        self.running_processes = {}
        self.set_up_scheduled_jobs()

        self.queue_thread = None
        self.schedule_thread = None
        if run_threads:
            self.start_threads()

    def start_threads(self):
        """A Scheduler instance needs two "worker threads"
        to operate. Each one runs in a loop forever.
        - one to process items in its action queue
        - one to add scheduled jobs to the action queue
          at the appropriate times

        This method creates and starts each of these threads.
        """
        self.queue_thread = threading.Thread(target=process_schedule, daemon=True)
        self.queue_thread.start()

        self.schedule_thread = threading.Thread(target=self.process_queue, daemon=True)
        self.schedule_thread.start()

    def set_up_scheduled_jobs(self):
        """Uses the `schedule` library to set up
        scheduled jobs to run on time.

        This is achieved by adding a RunFromSchedule
        action to the queue for each scheduled job at
        each scheduled time (could be once per day,
        once per hour, whatever the schedule defines).

        It's worth noting that it's probably easier
        not to mix scheduling and dependencies in the
        the same job. You can achieve the same thing
        by putting the schedule for the job in a separate
        "trigger" job, that becomes one of its upstream
        dependencies.
        """
        # NOTE: This method should never be run twice!
        #       If that happened, all scheduled jobs would
        #       be run twice...
        to_run = [t for t in self.jobs if t.schedule is not None]

        for job in to_run:
            logger.info(f"Setting up schedule for {job.name}")

            # Define a helper function which will
            # run the given job when called.
            def place_on_queue(job_to_run=job):
                logger.info(f"Running job {job_to_run.name} from its schedule")
                action = RunFromSchedule(job_to_run)
                self.queue.put(action)

            # We can now set up this helper function
            # to run on the defined schedule.
            job.schedule.do(place_on_queue)
            logger.info(f"Next run at {job.schedule.next_run}")

    def process_queue(self):
        logger.info("Processing queue...")
        while 1:
            try:
                self.process_one_queue_item()
            except StopProcessingQueueException:
                return

    def process_one_queue_item(self):
        """Inner function that is run in a tight loop in our
        `process_queue` function below. Setting this as its own
        standalone function makes unit testing a lot easier.
        """
        action = self.queue.get()
        logger.info("")
        logger.info(f"Got action from queue: {action}")

        job = action.job
        if job.status_counter != action.status_counter_snapshot:
            logger.info(
                f"The action {action} was created for job {job} when "
                f"that job had status_counter {action.status_counter_snapshot}, but by the time "
                f"the action reached the point of being carried out, the job "
                f"had a status_counter of {job.status_counter}. "
                f"Therefore the action will be aborted (not carried out)."
            )
            action.status = "aborted"
            return action

        method = {
            Reset: self.reset_job,
            Kill: self.kill_job,
            RunFromSchedule: self.run_job_from_schedule,
            Run: self.run_job,
            Spawn: self.spawn_job,
            RetryAfterDelay: self.retry_after_delay,
            RerunAfterDelay: self.retry_after_delay,
            SetCompletedSuccess: self.set_completed_success,
            SetCompletedFailure: self.set_completed_failure,
            StopProcessingQueueAction: self.stop_processing_queue,
        }[action.__class__]
        method(job, action)

        # Returning the action processed is helpful for unit testing
        return action

    def reset_job(self, job, action):
        # TODO: Check current status, decide if appropriate to continue
        job.update(JobStatuses.pending)
        action.status = "completed"

    def kill_job(self, job, action):
        # TODO: Check current status, decide if appropriate to continue
        if job.status in [
            JobStatuses.running,
            JobStatuses.pending_about_to_run,
        ]:
            # Kill the job by terminating its subprocess
            p = self.running_processes.get(job.name)
            if p is None:
                logger.error(f"Could not find a subprocess for {job.name}")
                action.status = f"aborted (could not find subprocess for {job.name}"
                return

            p.terminate()
            del self.running_processes[job.name]

            job.update(JobStatuses.killed)
            action.status = "completed"
        else:
            # If the job is not in one of the above
            # running-or-about-to-run states, the kill
            # action has no effect.
            action.status = f"aborted (cannot kill job in state {job.status})"

    def run_job_from_schedule(self, job, action):
        """This method is fired when a job reaches
        its scheduled time.

        We simply check if all its dependencies are
        successful, then if so we run the job.
        """
        if self.all_dependencies_succeeded(job):
            # This also works if the job has no dependencies
            logger.info(f"Running {job} on schedule, all dependencies completed")
            self.run_job(job, action)
        else:
            logger.info(f"Cannot run {job} on schedule, dependencies not completed")
            action.status = "aborted"

    def run_job(self, job, action):
        runnable = job.status in [
            JobStatuses.pending,
            JobStatuses.pending_waiting_to_retry,
            JobStatuses.killed,
            JobStatuses.failed,
            JobStatuses.failed_waiting_to_rerun,
            JobStatuses.succeeded,
        ]
        if not runnable:
            logger.info(
                f"Cannot run {job}, it is not in a runnable state ({job.status})"
            )
            action.status = "aborted (not in runnable state)"
            return

        job.update(JobStatuses.pending_searching_for_host)
        with self.hosts_lock:
            host = self.get_host_with_enough_ram(job)
            job.update_host(host)

        if host is None:
            job.update(JobStatuses.pending_no_host_found)
            logger.info(f"Could not find a host with enough free RAM for {job}")

            # We were unable to find an appropriate host with enough free RAM
            job.increment_failed_spawns()

            if job.failed_spawns <= len(self.retry_delays):
                delay = self.retry_delays[job.failed_spawns - 1]
                logger.info(f"Will retry running {job} in {delay} seconds")
                new_action = RetryAfterDelay(job, delay)
                self.queue.put(new_action)
                action.status = "aborted (no host found -> retry after delay)"

            else:
                logger.info(
                    f'Giving up on running {job}, placing it back into "waiting" state'
                )
                new_action = Reset(job)
                self.queue.put(new_action)
                action.status = "aborted (ho host found -> reset job)"

        else:
            job.update(JobStatuses.pending_about_to_run)

            new_action = Spawn(job)
            self.queue.put(new_action)
            action.status = "completed (-> spawn job)"

    def get_host_with_enough_ram(self, job):
        # TODO: We need some testing around this
        memory_usage = defaultdict(float)
        for t in self.jobs:
            if t.status in [
                JobStatuses.running,
                JobStatuses.pending_about_to_run,
            ]:
                memory_usage[t.host] += t.min_ram

        memory_free = {
            host: total - memory_usage.get(host, 0)
            for host, total in self.hosts.items()
        }

        candidates = [
            host for host, free_ram in memory_free.items() if free_ram >= job.min_ram
        ]

        return candidates[0] if candidates else None

    def spawn_job(self, job, action):
        """This is a rather hacky way of spawning
        a process. I don't like it much.

        But it guarantees that the job is spawned
        in its own process **with its own logger**.

        This means that jobs can log to files while
        the main scheduler process logs to stdout.

        It also means the job processes show up as
        their own separate processes in system tools
        such as `top` (their command will be the
        specific command for their file, it won't
        show up as `scheduler.py`).
        """
        # TODO: Check current status, decide if appropriate to continue
        assert job.host is not None

        logger.info(f"Spawning {job}")
        job.update(JobStatuses.running)

        # Spawn job in a thread to avoid blocking
        # main queue thread.
        future = self.run_job_subprocess(job, job.host)

        def on_completion(result):
            self.handle_completion(job, result._result, job.status_counter)

        future.add_done_callback(on_completion)
        action.status = "completed"

    def handle_completion(self, job, result, expected_status_counter):
        """
        This handler will be run when the job in question completes.

        We could simply set the completed status of the job here,
        in this function (which runs in some thread somewhere,
        separate from jobdec's queue of actions). But we prefer
        to use the queue, since it plays nicely with the general
        asynchronous approach we are using.
        """
        logger.info(f"{job} completed, result {result}")

        # Because there is necessarily a delay while the job for
        # this job runs, there is a chance that some other action
        # has been created for this job in the meantime.
        #
        # For example, the job might have been killed (or maybe
        # killed and re-run and it is now already running again).
        #
        # If this has happened, the job's status_counter will have
        # been incremented.
        #
        # If the status_counter is not what we expect, the job's result
        # should not be considered valid, and we abort.
        if job.status_counter != expected_status_counter:
            logger.warning(
                f"handle_completion expected {job} to have a status_counter of "
                f"{expected_status_counter}, but found a status_counter of {job.status_counter}. "
                f"Therefore the result of this job will be ignored."
            )
            # TODO: Could we update the job's action_history here somehow?
            return

        if result == JobStatuses.succeeded:
            new_action = SetCompletedSuccess(job)
        else:
            new_action = SetCompletedFailure(job)

        self.queue.put(new_action)

    @concurrent.thread  # in a concurrent thread so we don't delay scheduler's main loop with the below delay
    def retry_after_delay(self, job, action):
        job.update(JobStatuses.pending_waiting_to_retry)

        snapshot = job.status_counter
        time.sleep(action.delay)

        # Because there is a delay while the above sleep runs,
        # there is a chance that some other action has been
        # created for this job in the meantime.
        #
        # For example, the job might have been manually run.
        #
        # If this has happened, the job's status_counter will have
        # been incremented.
        #
        # If the status_counter is not what we expect, the job's result
        # should not be considered valid, and we abort. This is
        # a "manual" implementation of the logic which also resides
        # inside the action and Scheduler.process_queue, which is necessary
        # because otherwise this function could lead to a job being
        # run for a second time unintentionally.
        if job.status_counter != snapshot:
            logger.warning(
                f"run_after_delay expected {job} to have a status_counter of "
                f"{snapshot}, but found a status_counter of {job.status_counter}. "
                f"Therefore the run-after-delay will be aborted."
            )
            action.status = "aborted"
            return

        new_action = Run(job)
        self.queue.put(new_action)

        action.status = "completed"

    def set_completed_success(self, job, action):
        # As an absolutely minimal basic sanity check, the job
        # should be in running state if we are about to set it
        # to be completed.
        if job.status != JobStatuses.running:
            logger.warning(
                f"set_completed_success found {job} in status {job.status}, expected running"
            )
            action.status = "aborted"
            return

        with self.hosts_lock:
            job.update_host(None)

        job.update(JobStatuses.succeeded)

        to_run = []
        for name in self.graph.adj[job.name]:
            downstream = self.job_lookup[name]
            if self.all_dependencies_succeeded(downstream):
                to_run.append(downstream)

        for job in to_run:
            new_action = Run(job)
            self.queue.put(new_action)

        action.status = "completed"

    def set_completed_failure(self, job, action):
        # As an absolutely minimal basic sanity check, the job
        # should be in running state if we are about to set it
        # to be completed.
        if job.status != JobStatuses.running:
            logger.warning(
                f"set_completed_success found {job} in status {job.status}, expected running"
            )
            action.status = "aborted"
            return

        with self.hosts_lock:
            job.update_host(None)

        job.increment_failures()

        if job.failures <= len(job.rerun_delays):
            delay = job.rerun_delays[job.failures - 1]
            logger.info(f"Will retry running {job} in {delay} seconds")
            job.update(JobStatuses.failed_waiting_to_rerun)

            new_action = RerunAfterDelay(job, delay)
            self.queue.put(new_action)

            action.status = "completed (will rerun job after delay)"
        else:
            logger.info(f"Task {job} has failed, will not re-run it")
            job.update(JobStatuses.failed)

            action.status = "completed (job -> failed)"

    def stop_processing_queue(self, job, action):
        action.status = "completed"
        raise StopProcessingQueueException

    def all_dependencies_succeeded(self, job):
        dependency_jobs = [self.job_lookup[d] for d in job.depends_on]

        succeeded = [t.status == JobStatuses.succeeded for t in dependency_jobs]
        return all(succeeded)

    def get_job(self, job_name):
        return self.job_lookup[job_name]

    @concurrent.thread
    def run_job_subprocess(self, job, host):
        """This is a rather hacky way of spawning
        a process. I don't like it much.

        But it guarantees that the job is spawned
        in its own process **with its own logger**.

        This means that jobs can log to files while
        the main scheduler process logs to stdout.

        It also means the job processes show up as
        their own separate processes in system tools
        such as `top` (their command will be the
        specific command for their file, it won't
        show up as `scheduler.py`).
        """
        parts = job.name.split(".")
        module = ".".join(parts[:-1])
        func_name = parts[-1]

        # The python executable used here is `scheduler/venv/bin/python`
        command = [
            "python",
            "-u",
            "-c",
            f"from {module} import {func_name}; {func_name}()",
        ]

        # This bit is important - we can use the environment
        # variable in the subprocess to employ different logging.
        env = os.environ.copy()
        env["SCHEDULER_NAME"] = self.name
        env["SCHEDULER_TASK_NAME"] = job.name

        p = run_in_subprocess(command, env)

        # Assign this subprocess Popen to the Scheduler
        # object, so it can be terminated if necessary
        # (e.g. when `force_kill` is called).
        self.running_processes[job.name] = p

        p.wait()

        if p.returncode == 0:
            status = JobStatuses.succeeded
        else:
            status = JobStatuses.failed

        return status

    def force_kill(self, job):
        new_action = Kill(job)
        self.queue.put(new_action)
        return new_action

    def run_all_leaf_nodes(self):
        """Runs all jobs which do not depend on
        any other job (and are in pending state)
        """
        to_run = [
            t for t in self.jobs if not t.depends_on and t.status == JobStatuses.pending
        ]

        for job in to_run:
            action = Run(job)
            self.queue.put(action)


def run_in_subprocess(command, env):
    p = subprocess.Popen(command, env=env)
    return p


class StopProcessingQueueException(Exception):
    """A special exception used to break us out of the
    tight loop which processes the action queue. Should
    NEVER be raised in production, but can be useful to
    allow better control of the queue in unit tests.
    """


def process_schedule():
    while True:
        schedule.run_pending()
        time.sleep(1)


if __name__ == "__main__":
    logger.info(f"Main process PID: {os.getpid()}")
    logger.info(f"Main process name: {current_process().name}")
    logger.info(f"Main process logger: {id(logger)}")
    # # For a dead simple example
    # jobs = [Task('A', min_ram=4), Task('B', min_ram=5), Task('C', min_ram=3)]

    # # For a slightly more realistic example,
    # # importing jobs from the `jobs` folder,
    # # including dependency parsing.
    jobs_dir = full_path(__file__) / ".." / "jobs"
    collection = JobsCollection(jobs_dir)

    scheduler = Scheduler(
        "scheduler_main",
        collection.jobs.values(),
    )

    # This blocks forever while the Scheduler processes
    # items on its action queue (i.e. runs jobs according
    # to their schedules and dependencies).
    scheduler.queue_thread.join()



import os
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Union

from jobdec.file_utils import import_all_modules
from jobdec.scheduled_job import ScheduledJob


@dataclass
class Job:
    name: str
    schedule: str
    depends_on: tuple = tuple()


def get_dependency_string(upstream):
    if isinstance(upstream, str):
        return upstream
    if callable(upstream):
        return upstream.__module__ + "." + upstream.__name__
    raise ValueError(f"Dependency type not valid {type(upstream)}")


def get_module_string(job_name):
    """
    A job name is always, without exception,
    f.__module__ + '.' + f.__name__.

    So the module for a given job_name can
    be easily recreated.
    """
    out = ".".join(job_name.split(".")[:-1])
    return out


class GlobalJobsCollection:
    """
    This class's sole purpose is to collect
    all the job definitions from the `jobs`
    folder.

    These definitions can then be imported
    into the scheduler flexibly (not all
    collected jobs need necessarily be run
    by the scheduler).
    """

    jobs = {}

    def __call__(
        self,
        schedule=None,
        depends_on: tuple = (),
        minimum_ram: int = 1,
        rerun_delays: list = None,
    ):
        def decorator(f):
            job_name = f.__module__ + "." + f.__name__
            dependencies = tuple(get_dependency_string(other) for other in depends_on)
            job = ScheduledJob(
                name=job_name,
                schedule=schedule,
                function=f,
                depends_on=dependencies,
                min_ram=minimum_ram,
                rerun_delays=rerun_delays,
            )
            self.jobs[job_name] = job

            return f

        return decorator


class JobsCollection:
    def __init__(self, folder: Union[str, Path]):
        assert os.path.isdir(folder), f"{folder} is not a directory"

        # Import modules: this creates a Task for any function
        # decorated with the `@job` decorator, and places it into
        # the GlobalJobsCollection's class attribute `jobs`.
        module_strings = import_all_modules(folder)

        # However, the class attribute `jobs` dictionary therefore
        # simply grows and grows - it contains all job definitions
        # ever processed.
        # If a JobCollector is created with a path, we want it to
        # contain only jobs in the given folder.

        # TODO: Change this so that copies of Tasks are returned?
        #       I'm not sure if this is vital for now but it doesn't
        #       feel ideal using the same global Task objects in
        #       different JobsCollection objects.
        #       At the moment this is awkward because deepcopy
        #       uses pickle and we can't pickle thread.locks.
        self.jobs = {
            job_name: job
            for job_name, job in GlobalJobsCollection.jobs.items()
            if get_module_string(job_name) in module_strings
        }


GLOBAL_JOBS_COLLECTION = GlobalJobsCollection()

# Because it's neater to decorate a function with `@job`
# than `@GLOBAL_JOBS_COLLECTION`, we relabel.
job = GLOBAL_JOBS_COLLECTION



from datetime import datetime


class Action:
    def __init__(self, job):
        self.job = job
        self.status_snapshot = job.status
        self.status_counter_snapshot = job.status_counter
        self.follow_up_action = None

        self.status = "created"
        self.created_at = datetime.now()

        # No need to use job.lock here since appending to a list is thread safe, apparently
        self.job.action_list.append(self)

    def __str__(self):
        return f"Action: {self.__class__.__name__} {self.job}"

    def to_json(self):
        d = {
            "action": self.__class__.__name__,
            "status": self.status,
            "created_at": self.created_at.isoformat(),
        }
        return d


class Reset(Action):
    pass


class Kill(Action):
    # The action for a "force kill" operation. This results in
    # the same outcome as a Reset: the job should end up in
    # pending state. The only difference is that for a Kill
    pass


class ActionWithDelay(Action):
    def __init__(self, job, delay):
        super().__init__(job)
        self.delay = delay

    def __str__(self):
        return f"Action: {self.__class__.__name__}({self.delay}s) {self.job}"


class RetryAfterDelay(ActionWithDelay):
    pass


class RerunAfterDelay(ActionWithDelay):
    pass


class Run(Action):
    pass


class RunFromSchedule(Action):
    pass


class Spawn(Action):
    pass


class SetCompletedSuccess(Action):
    pass


class SetCompletedFailure(Action):
    pass


class StopProcessingQueueAction(Action):
    """An action which tells the loop which processes the action
    queue to exit
    """

    pass



import glob
import importlib
import os
from os.path import dirname, normpath, realpath, sep
from pathlib import Path
from typing import List, Union


def full_path(file) -> Path:
    dir = dirname(realpath(file))
    normalized = normpath(dir)
    out = Path(normalized)
    return out


def get_repo_directory():
    # NOTE: This assumes that the repo directory, the directory
    #       where all the code lives, is the first path in the
    #       PYTHONPATH environment variable.
    #       I think this is reasonable, since usually if we're
    #       trying to run our code, we'll want to the top-level
    #       repo directory in our PYTHONPATH. This just makes
    #       the small assumption that it is the first item.
    python_path = os.environ.get("PYTHONPATH")

    assert python_path, "PYTHONPATH env variable not set"
    repo_dir = python_path.split(":")[0]
    return repo_dir


def get_all_python_files_in(directory: Union[str, Path]) -> List[str]:
    """Returns the relative paths (relative to the
    top-level repo directory) of all Python files
    within the given directory.
    """
    assert sep == "/", f"This code only works on Linux ({sep})"
    directory = normpath(directory)

    directory = str(directory).rstrip(sep)
    paths = glob.glob(
        directory + "/**/*.py",
        recursive=True,
    )  # absolute paths

    # Convert to relative to top-level repo directory paths
    repo_dir = get_repo_directory()
    prefix = repo_dir + "/"
    assert all(x.startswith(prefix) for x in paths), f"Wrong prefixes: {paths}"
    n = len(prefix)
    relative_paths = [path[n:] for path in paths]
    relative_paths = [path.rstrip(sep) for path in relative_paths]  # just in case

    return relative_paths


def convert_to_import_strings(relative_filepaths: List[str]) -> List[str]:
    """Converts a list of Python filenames (specified
    relative to the top-level repo directory) into
    import strings (e.g. "talon.core.file_utils").
    """
    assert all(
        x.endswith(".py") for x in relative_filepaths
    ), f"Wrong suffixes: {relative_filepaths}"
    paths = [path[:-3] for path in relative_filepaths]

    import_strings = [path.replace(sep, ".") for path in paths]
    return import_strings


def import_all_modules(directory: Union[Path, str]) -> List[str]:
    """Returns all the modules within the given
    directory (i.e. Python module objects,
    imported using `importlib`).
    """
    import_strings = get_all_import_strings(directory)
    import_strings = sorted(import_strings)

    for x in import_strings:
        importlib.import_module(x)

    return import_strings


def get_all_import_strings(directory: Union[Path, str]) -> List[str]:
    python_files = get_all_python_files_in(directory)
    python_files = [f for f in python_files if not f.endswith("__init__.py")]
    import_strings = convert_to_import_strings(python_files)
    return import_strings



import datetime
import threading
from enum import Enum, auto

from jobdec.utils import logger

# User actions are Run, Kill
# These are the only two actions we need to care about interrupting other actions.
# We could also have SetCompleted actions appearing "out of nowhere", since these
# appear from a separate thread. However they should only ever appear when a job
# is in Running state.


class JobStatuses(Enum):
    pending = auto()
    pending_searching_for_host = auto()
    pending_no_host_found = auto()
    pending_about_to_run = auto()

    # This enum means "waiting to retry, having
    # failed to spawn the job due to not finding
    # a host with enough free memory".
    pending_waiting_to_retry = auto()

    running = auto()
    succeeded = auto()

    # This enum means "the job failed the first
    # time, but it will be re-run soon according
    # to its
    failed_waiting_to_rerun = auto()

    failed = auto()
    killed = auto()


class ScheduledJob:
    def __init__(
        self,
        name,
        function=None,
        depends_on=None,
        schedule=None,
        min_ram=1,
        rerun_delays=None,
    ):
        self.name = name
        self.function = function or (lambda: None)
        self.schedule = schedule
        self.depends_on = depends_on or []
        self.min_ram = min_ram

        self.status = JobStatuses.pending
        self.status_counter = 0
        self.last_updated = datetime.datetime.now().isoformat()

        self.lock = threading.Lock()
        self.failed_spawns = 0
        self.host = None
        self.action_list = []

        self.failures = 0
        self.rerun_delays = rerun_delays or []

    @property
    def next_run(self):
        return self.schedule.next_run if self.schedule else None

    def __str__(self):
        return f"Task({self.name}: {self.min_ram}MB)"

    def to_json(self):
        next_run = self.next_run
        scheduled_at = next_run.isoformat() if next_run else None
        d = {
            "name": self.name,
            "scheduled_at": scheduled_at,
            "depends_on": self.depends_on,
            "min_ram": self.min_ram,
            "status": self.status.name,
            "status_counter": self.status_counter,
            "last_updated": self.last_updated,
            "failed_spawns": self.failed_spawns,
            "host": str(self.host),
            "actions": [a.to_json() for a in self.action_list],
            "failures": self.failures,
            "rerun_delays": self.rerun_delays,
        }
        return d

    def update(self, new_status):
        with self.lock:
            self.status = new_status
            self.last_updated = datetime.datetime.now().isoformat()
            self.status_counter += 1
            logger.info(f"{self} -> {self.status.name} ({self.status_counter})")

    def increment_failed_spawns(self):
        with self.lock:
            self.failed_spawns += 1

    def increment_failures(self):
        with self.lock:
            self.failures += 1

    def update_host(self, host):
        self.host = host



import datetime
import time

import networkx as nx
from flask import Flask, Markup, jsonify, request

from jobdec.file_utils import full_path
from jobdec.graph import simple_jobs_to_networkx_graph
from jobdec.scheduled_job import JobStatuses


def create_app(scheduler):
    app = Flask(
        __name__,
        static_url_path="/static",
        static_folder=full_path(__file__) / "static",
    )
    app.scheduler = scheduler

    def health():
        pass

    def version():
        pass

    @app.route("/jobs")
    def jobs():
        # Returns jsonified details of all scheduler's jobs (jobs)
        return jsonify([t.to_json() for t in app.scheduler.jobs])

    @app.route("/jobs/<job_name>")
    def job(job_name):
        try:
            job = app.scheduler.get_job(job_name)
        except KeyError:
            resp = {"status": "failed", "detail": f"No job found named {job_name}"}
            return jsonify(resp)

        return jsonify(job.to_json())

    @app.route("/run_leaf_nodes")
    def run_leaf_nodes():
        # NOTE: This means "run all jobs with no upstream
        #       dependencies". It might be clearer to call
        #       this "run all root nodes"? Not sure.
        app.scheduler.run_all_leaf_nodes()
        return jsonify({"status": "success"})

    @app.route("/graph")
    def graph():
        # NOTE: This means "run all jobs with no upstream
        #       dependencies". It might be clearer to call
        #       this "run all root nodes"? Not sure.
        jobs = app.scheduler.job_lookup
        text = parse_jobs_to_svg(jobs)
        return Markup(text)

    @app.route("/kill/<job_name>")
    def kill(job_name):
        # NOTE: This means "run all jobs with no upstream
        #       dependencies". It might be clearer to call
        #       this "run all root nodes"? Not sure.
        try:
            job = app.scheduler.get_job(job_name)
        except KeyError:
            resp = {"status": "failed", "detail": f"No job found named {job_name}"}
            return jsonify(resp)

        action = app.scheduler.force_kill(job)

        # We wait to give the action a chance to
        # be processed. Yes, this is hacky, but
        # I think it's a lot more helpful to report
        # the result of the action here than to
        # simply report "success". Since we interact
        # with the Scheduler only by creating actions
        # I don't know of a clean way to specify
        # "only return the resulting action's status
        # when it has been processed".
        time.sleep(0.1)
        return jsonify({"status": action.status})

    def get_edges():
        pass

    def next_job():
        pass

    def running():
        # Returns jsonified details of all jobs (jobs) currently running
        pass

    def failed():
        # Returns jsonified details of all jobs (jobs) in failed state
        pass

    def status():
        pass

    def logs():
        pass

    def history():
        pass

    def force_kill():
        pass

    def force_run():
        pass

    def force_run_dependencies():
        pass

    def get_hosts():
        pass

    def update_hosts():
        pass

    return app


colours = {
    JobStatuses.pending: "#bbbbbb",
    JobStatuses.pending_searching_for_host: "#bbbbbb",
    JobStatuses.pending_no_host_found: "#bbbbbb",
    JobStatuses.pending_about_to_run: "#bbbbbb",
    JobStatuses.pending_waiting_to_retry: "#bbbbbb",
    JobStatuses.running: "#a0b3dd",
    JobStatuses.succeeded: "#03cc01",
    JobStatuses.failed_waiting_to_rerun: "#ff3233",
    JobStatuses.failed: "#ff3233",
    JobStatuses.killed: "#ff3233",
}


def parse_jobs_to_svg(jobs: dict):
    """
    Takes in a dictionary of {job_name: Task}
    and returns SVG text for displaying the given
    jobs on a web page.
    """
    g = simple_jobs_to_networkx_graph(list(jobs.values()))
    now = datetime.datetime.now()

    # Get a dictionary from job name to (x, y) position
    # positions = nx.nx_pydot.pydot_layout(g, prog='dot')
    # positions = nx.drawing.nx_agraph.graphviz_layout(g, prog='dot', args='-Grankdir=LR')
    # positions = nx.drawing.nx_agraph.graphviz_layout(g, prog='dot')
    positions = nx.drawing.nx_agraph.graphviz_layout(
        g, prog="dot", args="-Grankdir=BT -Granksep=1.5"
    )

    # Format here is start x, start y, width, height
    text = """
    <head>
      <link rel="stylesheet" type="text/css" href="/static/style.css">
      <link rel="preconnect" href="https://fonts.googleapis.com">
      <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
      <link href="https://fonts.googleapis.com/css2?family=Inter:wght@400;700&display=swap" rel="stylesheet">
    </head>
    """
    text += "<body>"

    # TODO: Adjust viewbox to size of graph
    text += """<svg viewBox="0 0 1600 900">"""

    for job_name, job in jobs.items():
        x, y = positions[job_name]

        final_name = job_name.split(".")[-1]
        short_name = final_name[0:16] + "..." if len(final_name) > 16 else final_name
        if (
            job.next_run
            and job.status == JobStatuses.pending
            and job.next_run - now < datetime.timedelta(days=1)
        ):
            next_run_str = job.next_run.time().isoformat()
            if next_run_str[-3:] == ":00":
                next_run_str = next_run_str[:-3]
            next_run_text = f"""
          <text class="info" font-family="Inter" font-size="11" fill="#555">
            <tspan x="102" y="63.8636">{next_run_str}</tspan>
          </text>
            """
        else:
            next_run_text = ""

        svg_class = " " + job.status.name

        # The rectangle starts at 10, 10, so if you line this up with
        # arrows remember to add 10 to the svg's x and y
        job_text = f"""
        <svg x={x} y={y} class="job{svg_class}" width="260" height="86" viewBox="0 0 260 86" fill="none" xmlns="http://www.w3.org/2000/svg">
          <rect x="10" y="10" width="239" height="65" rx="9" fill="white"/>
          <text class="title" fill="black" font-family="Inter" font-size="12" font-weight="bold">
            <tspan x="22" y="30.8636">
              <a href="jobs/{job_name}" target="_blank">{short_name}</a>
            </tspan>
          </text>
          <text class="info" fill="black" font-family="Inter" font-size="12">
            <tspan x="22" y="63.8636">{job.status.name}</tspan>
          </text>
          {next_run_text}
          <rect x="10" y="10" width="239" height="65" rx="9" stroke="#F0F0F0"/>
          <line class="line" x1="10" y1="42" x2="250" y2="42" stroke="#F0F0F0"/>
          <circle class="status" cx="211.5" cy="42.5" r="16" fill="white" stroke="#F0F0F0"/>
          <circle class="progress-circle" cx="211.5" cy="42.5" r="11.5" stroke="#06B6FF" stroke-width="3"/>
          <circle class="succeeded" cx="211.5" cy="42.5" r="7.5" fill="#FF9900"/>
          <circle class="failed" cx="211.5" cy="42.5" r="7.5" fill="#FF0000"/>
          <circle class="socket" cx="30" cy="10" r="4" fill="white" stroke="#F0F0F0"/>
          <circle class="socket" cx="30" cy="75" r="4" fill="white" stroke="#F0F0F0"/>
        </svg>
        """

        text += job_text

    text += """
      <!-- arrowhead marker definition -->
      <defs>
        <marker id="arrow" viewBox="0 0 10 10" refX="5" refY="5"
            markerWidth="6" markerHeight="6"
            orient="auto-start-reverse">
      <!-- fill matches stroke color-->
          <path d="M 0 0 L 10 5 L 0 10 z" fill="#666666" />
          <path d="M 0 0 L 10 5 L 0 10 z" fill="#666666" />
        </marker>
      </defs>
    """

    for e in g.edges:
        from_job_name, to_job_name = e
        from_x, from_y = positions[from_job_name]
        to_x, to_y = positions[to_job_name]
        arrow_text = f"""
          <polyline points="{from_x + 30},{from_y + 75} {to_x + 30},{to_y + 10}" fill="none" stroke="black" marker-end="url(#arrow)" />
        """
        text += arrow_text

    text += """</svg>"""
    text += "</body>"

    return text



import logging
import os


def set_up_logging():
    # create logger
    logger = logging.getLogger("__pyschedule__")
    logger.setLevel(logging.INFO)

    # create formatter
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    # We now decide what sort of logging to do,
    # based on whether we're in a scheduler job
    # process, or the main process.
    scheduler_name = os.environ.get("SCHEDULER_NAME")
    job_name = os.environ.get("SCHEDULER_TASK_NAME")
    if scheduler_name and job_name:
        # We're in a scheduler process, log to
        # an appropriate folder and file in /tmp
        folder = f"/tmp/{scheduler_name}"
        if not os.path.exists(folder):
            os.mkdir(folder)

        filename = f"{folder}/{job_name}.txt"
        handler = logging.FileHandler(filename)
    else:
        # We're in the main process, log to stdout
        handler = logging.StreamHandler()

    handler.setFormatter(formatter)

    logger.addHandler(handler)

    return logger


logger = set_up_logging()



from setuptools import setup

setup(
    name="jobdec",
    version="0.0.1",
    author="Gareth Williams",
    author_email="gareth.m.j.williams@gmail.com",
    packages=["jobdec"],
    package_data={"jobdec": ["static/*"]},
    url="http://pypi.python.org/pypi/jobdec/",
    license="LICENCE.txt",
    description="Creates and runs scheduled jobs from Python functions",
    long_description=open("README.md").read(),
    install_requires=[
        "Flask>=2",
        "networkx>=2",
        "Pebble>=4",
        "pygraphviz>=1",
        "schedule>=1",
    ],
)



import networkx as nx


class Job:
    def __init__(self, name, depends_on=None, produces=None, min_ram: int = 1):
        self.name = name
        self.depends_on = depends_on or []
        self.produces = produces or []
        self.min_ram = min_ram


class SimpleJob:
    def __init__(self, name, depends_on=None, min_ram: int = 1):
        self.name = name
        self.depends_on = depends_on or []
        self.min_ram = min_ram


def get_graph(jobs, edges):
    g = nx.DiGraph()
    for job in jobs:
        g.add_node(job.name)

    for e in edges:
        g.add_edge(*e)

    return g


def simple_jobs_to_networkx_graph(jobs: list):
    edges = [(dep, job.name) for job in jobs for dep in job.depends_on]

    g = get_graph(jobs, edges)
    return g


def jobs_to_networkx_graph(jobs: list):
    produced_by = {output: job.name for job in jobs for output in job.produces}

    edges = [(produced_by[dep], job.name) for job in jobs for dep in job.depends_on]

    g = get_graph(jobs, edges)
    return g


def get_node_positions(g) -> dict:
    return nx.drawing.nx_agraph.graphviz_layout(g, prog="dot", args="-Grankdir=LR")
